from PyQt5.QtWidgets import QGraphicsScene, QUndoStack, QUndoCommand
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QTransform
from effigy.QNodeSceneNode import QNodeSceneNode, QNodeSceneNodeUndeletable
from effigy.NodeIO import NodeIO, NodeLink

class NodeSceneModuleManager:
    def __init__(self):
        self.scene = None

    def selectNode(self, position, inType:type=None, outType:type=None):
        return None

class QNodeScene(QGraphicsScene):
    def __init__(self, moduleManager=None, *__args):
        self.undostack = QUndoStack()
        if moduleManager:
            self.moduleManager = moduleManager
        else:
            self.moduleManager = NodeSceneModuleManager()
        self.moduleManager.scene = self

        super().__init__(*__args)

        #self.mouseDoubleClickEvent.connect(self.mouseDoubleClickEvent)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_A:
            self.undostack.undo()
        elif event.key() == Qt.Key_D:
            self.undostack.redo()
        elif event.key() == Qt.Key_Delete:
            self.delSelection()

    def delSelection(self):


        selectedItems = self.selectedItems()

        links = [x for x in selectedItems if issubclass(type(x), NodeLink)]
        nodes = [x for x in selectedItems if issubclass(type(x), QNodeSceneNode) and not issubclass(type(x), QNodeSceneNodeUndeletable)]
        #remainder = [x for x in selectedItems if not issubclass(type(x), (QNodeSceneNode, NodeLink))]

        if links or nodes:
            self.undostack.beginMacro("Delete Stuff")
            for link in links:
                self.undostack.push(type(link.startIO).DeleteLinkCommand(link.startIO))

            for node in nodes:
                for io in node.IO.values():
                    self.undostack.push(type(io).DeleteLinkCommand(io))
                self.undostack.push(QNodeScene.DeleteNodeCommand(node))

            self.undostack.endMacro()

    class DeleteNodeCommand(QUndoCommand):
        def __init__(self, node):
            super().__init__()

            self.node = node
            self.scene = self.node.scene()

        def redo(self):
            self.scene.removeItem(self.node)

        def undo(self):
            self.scene.addItem(self.node)

    def mouseDoubleClickEvent(self, mouseEvent):
        if mouseEvent.button() == Qt.LeftButton and not self.itemAt(mouseEvent.scenePos(), QTransform()):
            self.undostack.beginMacro("Place Node")
            self.moduleManager.selectNode(mouseEvent.scenePos())    # Open Dialog for spawning new node
            self.undostack.endMacro()


from PyQt5.QtWidgets import QGraphicsItem, QUndoCommand
from PyQt5.QtCore import Qt

import uuid

from effigy.NodeIO import NodeIO, NodeLink


class QNodeSceneNode(QGraphicsItem):
    """Graphical representation of Node plus logic"""

    # Here we define some information about the node
    author = "DrLuke"       # Author of this node (only used for namespacing, never visible to users)
    modulename = "builtin"  # Internal name of the module, make this something distinguishable
    name = "Basenode"       # Human-readable name

    placeable = False       # Whether or not this node should be placeable from within the editor
    Category = ["Builtin"]     # Nested categories this Node should be sorted in

    # Description, similar to python docstrings (Brief summary on first line followed by a long description)
    description = """This node is the base class for all nodes.
It should never be placeable in the editor. However if you DO see this in the editor, something went wrong!"""

    def __init__(self, deserializeData=None, setID=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.IO = {}    # Stores all IO for this Node

        if setID is not None:
            self.id = setID  # Unique ID to identify node across multiple sessions.
        else:
            self.id = uuid.uuid4().int

        self.setFlag(QGraphicsItem.ItemIsMovable)   # Item can be dragged with left-click
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.setFlag(QGraphicsItem.ItemIsSelectable)

        self.addGraphicsItems()
        self.addIO()

        if deserializeData is not None:
            self.deserializeinternal(deserializeData)

    def serializeinternal(self):
        """Builtin internal node serializer, shouldn't be called from outside"""
        """
        {
            "pos": [float, float],
            "uuid": int,
            "type": string,
            "io": {
                "name1" : {
                    "links": [ [startid, endid], ... ] }
                ...
            "nodedata": {
                arbitrary stuff defined by the node itself
            }
        }
        """
        serdata = {}    # Contains serialized data of this object
        serdata["pos"] = [self.pos().x(), self.pos().y()]
        serdata["uuid"] = self.id
        serdata["modulename"] = type(self).modulename
        serdata["io"] = {}
        for io in self.IO.values():
            serdata["io"][io.id] = {}
            serdata["io"][io.id]["uuid"] = io.id
            serdata["io"][io.id]["name"] = io.name
            serdata["io"][io.id]["links"] = []
            for nodeLink in io.nodeLinks:
                serdata["io"][io.id]["links"].append([nodeLink.startIO.id, nodeLink.endIO.id, nodeLink.startIO.parentItem().id, nodeLink.endIO.parentItem().id])    #TODO: Add endio parent id?
        serdata["nodedata"] = self.serialize()

        return serdata

    def deserializeinternal(self, data):
        """Builtin internal node deserializer, shouldn't be called from outside"""
        self.setPos(data["pos"][0], data["pos"][1])

        # Reconstruct all
        sceneios = [x for x in self.scene().items() if issubclass(type(x), NodeIO)]
        for iodata in data["io"].values():
            if iodata["name"] in self.IO:
                self.IO[iodata["name"]].id = iodata["uuid"]

        for iodata in data["io"].values():
            for link in iodata["links"]:
                startio = None
                endio = None
                for io in sceneios:
                    if io.id == link[0]:
                        startio = io
                    if io.id == link[1]:
                        endio = io
                if startio is not None and endio is not None:
                    nl = NodeLink(startio, endio)
                    nl.updateBezier()
                    self.scene().addItem(nl)

        if data["nodedata"] is not None:
            self.deserialize(data["nodedata"])

    def serialize(self):
        """Custom node serializer, can be used to serialize non-default parameters"""
        raise NotImplementedError("This method must be implemented in derived class")

    def deserialize(self, data):
        """Custom node deserializer, restores node state with non-default parameters"""
        raise NotImplementedError("This method must be implemented in derived class")

    def addIO(self):
        """Add IOs to Node"""
        raise NotImplementedError("This method must be implemented in derived class")

    def addGraphicsItems(self):
        """Add Graphic Items for rendering the node"""
        raise NotImplementedError("This method must be implemented in derived class")

    def boundingRect(self):
        """Define appropriate bounding rect for Node body (necessary for interaction like dragging!)"""
        raise NotImplementedError("This method must be implemented in derived class")

    def selectedChanged(self, state):
        """React to being (de-)selected with visual cues"""
        raise NotImplementedError("This method must be implemented in derived class")

    def paint(self, *__args):
        pass

    class moveCommand(QUndoCommand):
        def __init__(self, item, newPos, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.selectedItems = set([x for x in item.scene().selectedItems() if issubclass(type(x), QNodeSceneNode)])
            self.positions = {}

            for iteritem in self.selectedItems:
                    self.positions[iteritem.id] = [iteritem, None, None]

            self.positions[item.id] = [item, item.pos(), newPos]

            self.firstredo = True

        def redo(self):
            if not self.firstredo:
                for position in self.positions.values():
                    if position[2] is not None:
                        position[0].setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)  # Disable, or it will generate a move command
                        position[0].setPos(position[2])
                        position[0].setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
                        position[0].updateLinkGraphics()  # Update links
            else:
                self.firstredo = False  # Workaround, because the action already happens somewhere else when the command is created

        def undo(self):
            for position in self.positions.values():
                if position[1] is not None:
                    position[0].setFlag(QGraphicsItem.ItemSendsGeometryChanges, False)  # Disable, or it will generate a move command
                    position[0].setPos(position[1])
                    position[0].setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
                    position[0].updateLinkGraphics()  # Update links

        def id(self):
            return 1

        def mergeWith(self, command):
            # Merge command with this
            # Note: 'command' is a newer command than this
            if command.selectedItems == self.selectedItems:
                for item in self.selectedItems:
                    # Take start position from older command, and new position from newer command if it isn't None
                    oldpos = self.positions[item.id][1] if self.positions[item.id][1] is not None else command.positions[item.id][1]
                    # Take end position from newest command
                    newpos = command.positions[item.id][2] if command.positions[item.id][2] is not None else self.positions[item.id][2]
                    self.positions[item.id] = [self.positions[item.id][0], oldpos, newpos]
                return True     # Indicate success
            else:
                return False

    def updateLinkGraphics(self):
        for IO in self.IO.values():
            for link in IO.nodeLinks:
                link.updateBezier()

    def mouseMoveEvent(self, QGraphicsSceneMouseEvent):
        super().mouseMoveEvent(QGraphicsSceneMouseEvent)

    def mousePressEvent(self, QGraphicsSceneMouseEvent):
        selectedNodes = [x for x in self.scene().selectedItems() if issubclass(type(x), QNodeSceneNode)]
        if QGraphicsSceneMouseEvent.button() == Qt.LeftButton and len(selectedNodes) > 1:
            self.scene().undostack.beginMacro("Move multiple nodes")
        super().mousePressEvent(QGraphicsSceneMouseEvent)

    def mouseReleaseEvent(self, QGraphicsSceneMouseEvent):
        selectedNodes = [x for x in self.scene().selectedItems() if issubclass(type(x), QNodeSceneNode)]
        if QGraphicsSceneMouseEvent.button() == Qt.LeftButton and len(selectedNodes) > 1:
            self.scene().undostack.endMacro()
        super().mouseReleaseEvent(QGraphicsSceneMouseEvent)

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            move = QNodeSceneNode.moveCommand(self, value)
            self.scene().undostack.push(move)

        elif change == QGraphicsItem.ItemPositionHasChanged:
            self.updateLinkGraphics()

        elif change == QGraphicsItem.ItemSelectedHasChanged:
            self.selectedChanged(value)

        return super().itemChange(change, value)


class QNodeSceneNodeUndeletable(QNodeSceneNode):
    """Nodes that can't be deleted from a sheet"""
    pass


from PyQt5.QtWidgets import QGraphicsItem

from PyQt5.QtWidgets import QGraphicsRectItem, QGraphicsPathItem, QUndoCommand, QGraphicsItem
from PyQt5.QtCore import QRectF, QPointF, Qt
from PyQt5.QtGui import QTransform, QBrush, QColor, QPainterPath, QPen
from PyQt5.QtWidgets import QUndoCommand


from enum import Enum
import uuid

#from effigy.NodeLink import NodeLink, InvalidLinkException

class NodeIODirection(Enum):
    """Establish if IO direction is input, output or any"""
    any = 0
    output = 1
    input = 2

class NodeIOMultiplicity(Enum):
    """Whether the node can be connected to multiple other nodes, or just one"""
    single = 0
    multiple = 1

class NodeIO(QGraphicsItem):
    """General Input and Output class for Node links.

    This class builds the basis for inputs and outputs for nodes. You link nodes together via those inputs and outputs,
    and all IO must be typed. Only compatible IO (that is, IO of same type) can be linked together. Compatibility is
    established with the 'isinstance' method. Each IO has a direction, i.e. input, output or any. Inputs may only be
    conected with inputs (and 'any' of course), and vice versa."""

    # Direction for this class
    classDirection = NodeIODirection.any
    # Multiplicity for this class
    classMultiplicity = NodeIOMultiplicity.multiple


    def __init__(self, iotype, name, displaystr=None, setID=None, *args, **kwargs):
        if type(iotype) is not type and type(iotype) is not list:
            raise TypeError("iotype argument is of type '%s', but must be of type 'type' or 'list'." % type(iotype))

        super().__init__(*args, **kwargs)

        self.iotype = iotype
        self.iodirection = NodeIO.classDirection

        if setID is not None:
            self.id = setID  # Unique ID to identify io across multiple sessions.
        else:
            self.id = uuid.uuid4().int

        self.name = str(name)   # Each IO needs a name unique across all instances of the node class using it
        self.displaystring = displaystr     # Preferred String to display next to IO when rendering. Not required for it to work.

        self.addGraphicsItems()

        self.nodeLinks = []
        self.tempNodeLink = None

    # Define a bounding rect for collision detection (needed for interacting with the IO)
    def boundingRect(self):
        return self.backgroundItem.rect()

    def addGraphicsItems(self):
        """Add Items to draw the IO pin. Can be overridden for custom graphics"""
        self.backgroundItem = QGraphicsRectItem(QRectF(-5, -5, 10, 10), self)

    def paint(self, *__args):
        """Must be defined, but as this is just a container item, it doesn't actually paint anything."""
        pass

    def deleteLink(self, link):
        link.startIO.nodeLinks.remove(link)
        link.endIO.nodeLinks.remove(link)
        self.scene().removeItem(link)

    def recursiveIOCheck(self, item):
        """Recursively check if item is a child of an IO item"""
        if item is None:  # We reached the topmost item
            return None
        if isinstance(item, NodeIO):  # Found a NodeIO
            return item
        else:
            return self.recursiveIOCheck(item.parentItem())  # We must go deeper and deeper!

    class DeleteLinkCommand(QUndoCommand):
        def __init__(self, io):
            super().__init__()
            self.io = io
            self.linkIOs = []
            self.scene = io.scene()

            for link in io.nodeLinks:
                self.linkIOs.append([link.startIO, link.endIO, link])

        def redo(self):
            for linkIO in self.linkIOs:
                linkIO[0].deleteLink(linkIO[2])

        def undo(self):
            for linkIO in self.linkIOs:
                linkIO[0].nodeLinks.append(linkIO[2])
                linkIO[1].nodeLinks.append(linkIO[2])

                self.scene.addItem(linkIO[2])


    class CreateLinkCommand(QUndoCommand):
        def __init__(self, startIO, endIO, scene, *args, **kwargs):
            self.startIO = startIO
            self.endIO = endIO
            self.scene = scene

            self.link = NodeLink(startIO=self.startIO, endIO=self.endIO, undostack=self.scene.undostack)

            self.firstredo = True

            super().__init__(*args, **kwargs)

        def redo(self):
            if not self.firstredo:
                self.link.startIO.nodeLinks.append(self.link)
                self.link.endIO.nodeLinks.append(self.link)
            else:
                self.firstredo = False

            self.link.updateBezier()
            self.scene.addItem(self.link)

        def undo(self):
            self.startIO.deleteLink(self.link)

    def mousePressEvent(self, QGraphicsSceneMouseEvent):
        if QGraphicsSceneMouseEvent.button() == Qt.LeftButton:
            self.scene().undostack.beginMacro("Link Nodes")
            if type(self).classMultiplicity == NodeIOMultiplicity.single:
                if self.nodeLinks:
                    self.scene().undostack.push(NodeIO.DeleteLinkCommand(self))

            if self.tempNodeLink is None:
                if self.classDirection == NodeIODirection.input:
                    # On inputs, io is the ending point of bezier curve
                    self.tempNodeLink = NodeLink(startIO=None, endIO=self)
                else:
                    # Everywhere else, io is start point of bezier curve
                    self.tempNodeLink = NodeLink(startIO=self, endIO=None)
                self.scene().addItem(self.tempNodeLink)


    def mouseMoveEvent(self, QGraphicsSceneMouseEvent):
        # update temporary link
        if self.tempNodeLink is not None:
            if self.classDirection == NodeIODirection.input:
                self.tempNodeLink.updateBezier(overrideStartpos=QGraphicsSceneMouseEvent.scenePos())
            else:
                self.tempNodeLink.updateBezier(overrideEndpos=QGraphicsSceneMouseEvent.scenePos())

    def mouseReleaseEvent(self, QGraphicsSceneMouseEvent):
        if QGraphicsSceneMouseEvent.button() == Qt.LeftButton:
            if self.tempNodeLink is None:
                return

            #  Delete temporary link
            self.scene().removeItem(self.tempNodeLink)
            self.tempNodeLink = None

            # Try to find any item at end coordinates
            endpos = QGraphicsSceneMouseEvent.scenePos()
            itematendpos = self.scene().itemAt(endpos, QTransform())

            # Check if item is nodeIO
            enditem = self.recursiveIOCheck(itematendpos)
            if enditem is not None:
                try:
                    # Check if IO directions are compatible
                    if self.classDirection == enditem.classDirection and not self.classDirection == NodeIODirection.any and not enditem.classDirection == NodeIODirection.any:
                        raise InvalidLinkException
                    if self.classDirection == NodeIODirection.input:
                        # On inputs, io is the ending point of bezier curve
                        linkAction = NodeIO.CreateLinkCommand(startIO=enditem, endIO=self, scene=self.scene())
                        if linkAction.link.linkgood:
                            self.scene().undostack.push(linkAction)
                    else:
                        # Everywhere else, io is start point of bezier curve
                        linkAction = NodeIO.CreateLinkCommand(startIO=self, endIO=enditem, scene=self.scene())
                        if linkAction.link.linkgood:
                            self.scene().undostack.push(linkAction)
                except InvalidLinkException:
                    pass

            else:   # No node found. Launch the module selector to spawn a compatible node
                self.scene().undostack.beginMacro("Create node from link")
                if self.classDirection == NodeIODirection.input:
                    returnio = self.scene().moduleManager.selectNode(QGraphicsSceneMouseEvent.pos(), inType=self.iotype)
                    if issubclass(type(returnio), NodeIO):
                        linkAction = NodeIO.CreateLinkCommand(startIO=returnio, endIO=self, scene=self.scene())
                        if linkAction.link.linkgood:
                            self.scene().undostack.push(linkAction)
                else:
                    returnio = self.scene().moduleManager.selectNode(QGraphicsSceneMouseEvent.pos(), outType=self.iotype)
                    if issubclass(type(returnio), NodeIO):
                        linkAction = NodeIO.CreateLinkCommand(startIO=self, endIO=returnio, scene=self.scene())
                        if linkAction.link.linkgood:
                            self.scene().undostack.push(linkAction)
                self.scene().undostack.endMacro()
            self.scene().undostack.endMacro()
        elif QGraphicsSceneMouseEvent.button() == Qt.RightButton:
            self.scene().undostack.push(NodeIO.DeleteLinkCommand(self))




class NodeInput(NodeIO):
    """General Input class for Node links. See NodeIO class for more information."""
    classDirection = NodeIODirection.input
    classMultiplicity = NodeIOMultiplicity.single

class NodeOutput(NodeIO):
    """General Output class for Node links. See NodeIO class for more information."""
    classDirection = NodeIODirection.output
    classMultiplicity = NodeIOMultiplicity.multiple


class InvalidLinkException(Exception):
    pass


class DeleteLinkFromMultiplicityCommand(QUndoCommand):
    """Only use in combination of the "Link IO" macro from NodeIO"""
    def __init__(self, link):
        super().__init__()
        self.link = link
        self.scene = self.link.scene()

    def redo(self):
        self.link.startIO.nodeLinks.remove(self.link)
        self.link.endIO.nodeLinks.remove(self.link)
        self.scene.removeItem(self.link)

    def undo(self):
        self.link.startIO.nodeLinks.append(self.link)
        self.link.endIO.nodeLinks.append(self.link)
        self.scene.addItem(self.link)

class NodeLink(QGraphicsPathItem):
    def __init__(self, startIO=None, endIO=None, undostack=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.startIO = startIO
        self.endIO = endIO
        self.undostack = undostack

        self.linkgood = False

        # Connections with yourself are not allowed
        if startIO == endIO:
            raise InvalidLinkException("startIO (ID: %s) and endIO (ID: %s) can not be the same!" % (startIO.id, endIO.id))

        if startIO is not None and endIO is not None:
            # Start type can be casted to end type
            typesCompatible = False
            if type(endIO.iotype) is list:
                if len(endIO.iotype) == 0:
                    typesCompatible = True
                else:
                    for iotype in endIO.iotype:
                        if issubclass(startIO.iotype, iotype):
                            typesCompatible = True
                            break

            else:
                typesCompatible = issubclass(startIO.iotype, endIO.iotype)

            if typesCompatible:
                self.linkgood = True
                # Remove duplicates
                for link in self.startIO.nodeLinks:
                    if link.startIO == self.startIO and link.endIO == self.endIO:
                        self.startIO.nodeLinks.remove(link)

                # Remove all other links if multiplicity calls for it
                if type(self.startIO).classMultiplicity == NodeIOMultiplicity.single:
                    for link in self.startIO.nodeLinks:
                        self.undostack.push(DeleteLinkFromMultiplicityCommand(link))
                self.startIO.nodeLinks.append(self)

                # Remove all other links if multiplicity calls for it
                if type(self.endIO).classMultiplicity == NodeIOMultiplicity.single:
                    for link in self.endIO.nodeLinks:
                        self.undostack.push(DeleteLinkFromMultiplicityCommand(link))
                self.endIO.nodeLinks.append(self)

    def updateBezier(self, overrideStartpos=QPointF(0, 0), overrideEndpos=QPointF(0, 0)):
        path = QPainterPath()

        if self.startIO is not None:
            startpos = self.startIO.scenePos()
        else:
            startpos = overrideStartpos

        if self.endIO is not None:
            endpos = self.endIO.scenePos()
        else:
            endpos = overrideEndpos

        path.moveTo(startpos)

        controlpoint = QPointF(abs((endpos - startpos).x()) * 0.8, 0)
        path.cubicTo(startpos + controlpoint,
                     endpos - controlpoint,
                     endpos)

        self.setPath(path)

    def paint(self, painter, option, widget=None):
        if self.isSelected():
            self.setPen(QPen(Qt.red))
        else:
            self.setPen(QPen(Qt.black))
        painter.setPen(self.pen())
        painter.drawPath(self.path())
